#!/usr/bin/python3

# @package      hubzero-mw2-front-proxy
# @file         front-proxy.py
# @copyright    Copyright (c) 2016-2020 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
# HUBzero Front Proxy
#
#
# This proxy runs on the web front-end and execution hosts and handles AJAX 
# and WebSocket connections.  Apache 2.4 or better can be used to redirect 
# AJAX and WSS into a locally-running copy of this proxy.  Earlier versions 
# of Apache cannot act as a reverse proxy for WS/WSS, so this proxy must run 
# on a separate IP address.
#
# This proxy forks a child process for each connection, interprets the
# HTTP header, looks up the information in a database, and chooses a
# host:port to forward the connection to.
#
# This version of the proxy encompasses both the earlier hubzero-mw2-front-proxy
# and hubzero-mw2-exec-proxy functions with the following enhancements:
#
# - Python 3
# - Uses python ssl module instead of pyOpenSSL
# - Supports persistent HTTP connections
#

import pprint
import socket
import time
import sys
import os
import errno
import traceback
import fcntl
import syslog
import select
import signal

try:
    import MySQLdb
except:
    import pymysql
    pymysql.install_as_MySQLdb()
    import MySQLdb

import logging
import grp
import pwd
import argparse
import ssl
try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser as ConfigParser
import simplejson
import yaml

def fatal(str):

    if os.geteuid() == 0:
        syslog.syslog('FATAL: ' + str)
    else:
        logging.critical('FATAL: ' + str)

    # Note that we use sys.exit here to throw a SystemExit up the call stack.

    sys.exit(1)


def log_exception(title):

    cla, exc, trbk = sys.exc_info()

    excName = cla.__name__

    logging.error(title + ' ' + excName + ':')
    logging.error(str(exc))
    logging.error('-' * 50)

    excTb = traceback.format_tb(trbk, 10)

    for entry in excTb:
        for line in entry.strip().split('\n'):
            logging.error(line.strip())

    logging.error('-' * 50)


# ==============================================================================
# Enable/disable blocking for a SSL, socket, or file descriptor.
# ==============================================================================

def setblocking(x, value):

    if 'ssl' in str(type(x)):
        x.setblocking(value)
        return

    if 'socket' in str(type(x)):
        # Replace x with the file descriptor and fall through to cases below.
        x = x.fileno()

    if not isinstance(x, type(1)):
        fatal("Don't know how to set blocking of type %s" % str(type(x)))

    if value:
        # Do block on I/O.
        fcntl.fcntl(x, fcntl.F_SETFL, fcntl.fcntl(x, fcntl.F_GETFL) & ~os.O_NONBLOCK)
    else:
        # Do not block on I/O.
        fcntl.fcntl(x, fcntl.F_SETFL, fcntl.fcntl(x, fcntl.F_GETFL) | os.O_NONBLOCK)


# ==============================================================================
# Occasionally flush old entries from the cache.
# ==============================================================================

def flush_cache(cache, t):

    if t - int(t) < cfg.CACHE_FLUSH_PERCENT / 100.0:

        sessions = list(cache)

        for sessnum in sessions:
            if cache[sessnum][4] + cfg.CACHE_DURATION < t:
                del cache[sessnum]


# ==============================================================================
# Try to read all pipes open to all child processes.  Get the session number,
# username, filexfer cookie, execution host, and container number.  Incorporate
# that information into the cache---as well as the time at which it occurred.
# Close the pipe reader when it has been read.
# ==============================================================================

def read_cache_pipes(rlist, cache, t):

    for r in rlist:
        try:
            stuff = os.read(r, 1000)

            os.close(r)

            rlist.remove(r)

            if stuff == b'':
                continue

            try:
                sessnum, username, cookie, exechost, ctnum = stuff.split(b',')
            except BaseException:
                logging.warning("read_cache_pipes(): Unable to split cache info '%s' from %s" % (stuff, r))
                continue

            try:
                sessnum = int(sessnum)
                ctnum = int(ctnum)
            except BaseException:
                logging.warning("Bad format for cache values")
                continue

            cache[sessnum] = (username.decode('utf_8'), cookie.decode('utf_8'), exechost.decode('utf_8'), ctnum, t)

        except OSError as error:

            if error.errno == errno.EAGAIN:
                pass
            else:
                fatal("read_cache_pipes(): Error %d in read pipe %s" % (error.errno, r))


# ==============================================================================
# Write the sessnum, username, filexfer cookie, execution host and container
# number to the pipe open with the parent.
# ==============================================================================

def write_cache_pipe(w, sessnum, username, cookie, host, ct):

    # A write to the pipe might fail if the reader already closed it.
    # This could happen if the parent process was killed.

    cacheline = "%d,%s,%s,%s,%d" % (sessnum, username, cookie, host, ct)

    try:
        os.write(w, cacheline.encode('utf_8'))
    except BaseException as e:
        logging.warning("write_cache_pipe() unable to write cacheline to parent[%s]: %s" % (w,str(e)))
        pass

    os.close(w)


# ==============================================================================
# Create a socket to listen on.  Optionally wrap it with SSL.
# ==============================================================================

def create_listener(host, port):

    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(5)

    return sock

def create_ssl_context():

    if cfg.SSL_PRIVKEY and os.path.isfile(cfg.SSL_PRIVKEY):
        front_key = cfg.SSL_PRIVKEY
        front_cert = cfg.SSL_CERT
        front_cacert = cfg.SSL_CACERT
    elif cfg.SSL_PRIVKEY:
        front_key = os.path.join(cfg.CONFIG_DIR, cfg.SSL_PRIVKEY)
        front_cert = os.path.join(cfg.CONFIG_DIR, cfg.SSL_CERT)

        if cfg.SSL_CACERT is not None:
            front_cacert = os.path.join(cfg.CONFIG_DIR, cfg.SSL_CACERT)
        else:
            front_cacert = None

    if hasattr(ssl, "PROTOCOL_TLSv1_2"):
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        ctx.options |= ssl.OP_NO_TLSv1
        ctx.options |= ssl.OP_NO_TLSv1_1
    else:
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1)

    ctx.options |= ssl.OP_NO_SSLv2
    ctx.options |= ssl.OP_NO_SSLv3

    ctx.load_cert_chain(certfile=front_cert, keyfile=front_key)

    if front_cacert is not None and front_cacert != '':
        ctx.load_verify_locations(front_cacert)

    if cfg.SSL_DHPARAM is not None:
        ctx.load_dh_params(os.path.join(cfg.CONFIG_DIR, cfg.SSL_DHPARAM))

    if cfg.SSL_CIPHERS is not None:
        ctx.set_ciphers(cfg.SSL_CIPHERS)

    ctx.options |= ssl.OP_SINGLE_DH_USE

    return ctx


# ==============================================================================
# Exceptions for reading/writing.
# ==============================================================================

class ReaderClose(Exception):
    """Socket or SSL is closed for reading."""
    pass


class WriterClose(Exception):
    """Socket or SSL is closed for writing."""
    pass


class Timeout(Exception):
    """Socket I/O timeout timer expired"""
    pass


# ==============================================================================
# Read a string from a socket.
# ==============================================================================

def read_chunk(s, maxlen=65536):
    host, port = s.getpeername()
    
    try:
        msg = s.recv(maxlen)
    except ssl.SSLZeroReturnError:
        raise ReaderClose()
    except Timeout:
        raise Timeout()
    except socket.error as error:
        if error.errno == errno.EWOULDBLOCK:
            return b''
        if error.errno == errno.EAGAIN:
            return b''
        elif error.errno == 2: # SSLWantReadError
            # See SSLWantReadError below
            #logging.debug("read_chunk(%s:%d): [SSLWantReadError:%d] %s (continue)" % (host, port, error.errno, error))
            return b''
        elif error.errno == 3: # SSLWantWriteError
            logging.debug("read_chunk(%s:%d): [SSLWantWriteError] %s (continue)" % (host, port, error))
            return b''
        else:
            logging.debug("read_chunk(%s:%d): Caught socket.error %d %s (close connection)" % (host, port, error.errno, str(error)))
            raise ReaderClose()
    except ssl.SSLWantReadError as errors:
        # In particular when running in non-blocking mode the socket could be in the middle of
        # an SSL protocol transaction which triggered a select(). In this case the
        # read should be repeated until the SSL protocol transaction is completed and
        # user data is actually available. We won't log this as it is expected behavior.
        #logging.debug("read_chunk(%s:%d): Caught ssl.SSLWantReadError %s (continue)" % (host, port, str(errors)))
        return b''
    except ssl.SSLWantWriteError as errors:
        # In particular when running in non-blocking mode the socket could be in the middle of
        # an SSL protocol transaction which triggered a select(). In this case the
        # read should be repeated until the SSL protocol transaction is completed and
        # user data is actually available. We won't log this as it is expected behavior.
        logging.debug("read_chunk(%s:%d): Caught ssl.SSLWantWriteError %s (continue)" % (host, port, str(errors)))
        return b''
    except ssl.SSLError as errors:
        if 'wrong version number' in str(errors):
            logging.debug("read_chunk(%s:%d): invalid ssl handshake (close connection)" % (host, port))
        # Some clients seem to drop connection right after 404 result which causes an ssl.SSLError Unexpected EOF.
        # We won't log this as it is expected behavior. This can also occur when a client abruptly closes
        # websocket connection. ECONNRESET can also occur with some clients in similar conditions.
        elif 'Unexpected EOF' not in str(errors) and 'ECONNRESET' not in str(errors):
            logging.debug("read_chunk(%s:%d): Caught ssl.SSLError: %s (close connection)" % (host, port, str(errors)))
        raise ReaderClose()
    except OSError as error:
        if error.errno == errno.EAGAIN:
            logging.debug('read_chunk(%s:%d): Caught EAGAIN (continue)' % (host, port))
            return b''
        else:
            logging.debug('read_chunk(%s:%d): Caught errno = %d (close connection)' % (host, port, error.errno))
            raise ReaderClose()

    # If socket peer has performed an orderly shutdown then we get an empty string.

    if msg == b'' and maxlen > 0:
        raise ReaderClose()

    return msg


# ==============================================================================
# Write a string to a socket.  Return the unwritten part of the string.
# ==============================================================================

def write_chunk(s, msg):
  
    # if the client dropped connection before this is called then
    # getpeername() fails. it would be better if we captured the host
    # and port early on in the process and kept it globally rather
    # than attempt to fetch it again everywhere with getpeername()
    # for now we move this one into the try so we can catch the error

    host = 'unknown'
    port = -1

    try:
        host, port = s.getpeername()
        n = s.send(msg)
    except OSError as error:
        if error.errno == 3: # SSLWantWriteError
            # See SSLWantWriteError below
            #logging.debug("write_chunk(%s:%d): [SSLWantWriteError:%d] %s (continue)" % (host, port, error.errno, error))
            return  msg
        elif error.errno == errno.EAGAIN:
            logging.debug("write_chunk(%s:%d): EAGAIN (continue)" % (host, port))
            return  msg
        else:
            fatal("write_chunk(%s:%d) Error %d in write_chunk: %s" % (host, port, error.errno, error))                
    except ssl.SSLWantWriteError:
        # In particular when running in non-blocking mode the socket could be in the middle of
        # an SSL protocol transaction which triggered a select(). In this case the
        # write should be repeated until the SSL protocol transaction is completed and
        # user data can actually be written. We won't log this as it is expected behavior.
        #logging.debug("write_chunk(%s:%d): ssl.SSLWantWriteError (continue)" % (host, port))
        return msg
    except (ssl.SSLWantReadError):
        # In particular when running in non-blocking mode the socket could be in the middle of
        # an SSL protocol transaction which triggered a select(). In this case the
        # write should be repeated until the SSL protocol transaction is completed and
        # user data can actually be written. We won't log this as it is expected behavior.
        logging.debug("write_chunk(%s:%d): ssl.SSLWantReadError (continue)" % (host, port))
        return msg
    except ssl.SSLZeroReturnError:
        fatal("write_chunk(%s:%d): Caught ssl.SSLZeroReturnError" % (host, port))
    except ssl.SSLError as errors:
        fatal("write_chunk(%s:%d): Caught ssl.SSLError: %s" % (host, port, str(errors)))
    except socket.error as errors:
        fatal("write_chunk(%s:%d): Caught socket.error: %s" % (host, port, str(errors)))

    return msg[n:]


# ==============================================================================
# Send a shutdown only for an SSL socket.  Then close it.
# ==============================================================================

def shutdown(s):

    setblocking(s, True)

    try:
        s.shutdown(socket.SHUT_RDWR)
    except BaseException as e:  # Could be that the connection is already shutdown
        # EPIPE could occur on an SSL socket shutdown if it tries to send SSL
        # protocol shutdown data to a socket that is already closed. We'll
        # consider this a possible normal condition and not log the error.
        if 'EPIPE' not in str(e):
            logging.warning("shutdown(): unable to shutdown cleanly %s" % str(e))

    s.close()


# ==============================================================================
# Read an HTTP header, return it as an array.
# Any additional data read after the double CRLF is returned as body.
# ==============================================================================

def read_header(ns, timeout = 0):
    global debug_header

    chunk = b''

    if (timeout > 0):
        signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(cfg.HEADER_TIMEOUT)

    while chunk.rfind(b'\r\n\r\n') == -1:

        first = (len(chunk) == 0)

        try:
            chunk += read_chunk(ns)
        except ReaderClose:
            break

        if first:
            debug_header = chunk.split(b'\r')[0]

        if len(chunk) > 100000:
            fatal('read_header(): Header is too long (%d)' % (len(chunk)))

    if (timeout > 0):
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

    arr = chunk.split(b'\r\n\r\n', 1)

    hdr = arr[0].split(b'\n')

    for n in range(0, len(hdr)):
        hdr[n] = hdr[n].strip().decode('utf_8')

    if len(hdr) < 1:
        fatal('read_header(): Malformed header: ' + str(hdr))

    if len(arr) == 2:
        body = arr[1]
    else:
        body = b''

    return hdr, body


# ==============================================================================
# Send the header through a socket/SSL.
# ==============================================================================

def send_header(s, hdr, timeout = 0):

    chunk = '\r\n'.join(hdr).encode('utf_8') + b'\r\n\r\n'

    if (timeout > 0):
        signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(cfg.HEADER_TIMEOUT)

    while len(chunk) > 0:
        chunk = write_chunk(s, chunk)

    if (timeout > 0):
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

    if 0 and cfg.VERBOSITY > 1:

        i = header_find(hdr, 'Connection:')

        if i < 0:
            fatal("send_header(): No Connection: statement")

        if hdr[i] != 'Connection: close':
            logging.debug('send_header(): >>>>> ' + hdr[i])


# ==============================================================================
# Send data through a socket/SSL.
# ==============================================================================

def send_data(s, data, timeout = 0):

    chunk = data

    if (timeout > 0):
        signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(cfg.HEADER_TIMEOUT)

    while len(chunk) > 0:
        chunk = write_chunk(s, chunk)

    if (timeout > 0):
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

# ==============================================================================
# Log the full HTTP header.
# ==============================================================================

def log_header(hdr):
    logging.debug('================================')

    for h in hdr:
        logging.debug(h)


# ==============================================================================
# Find an entry in the header.
# ==============================================================================

def header_find(hdr, s):

    for n in range(0, len(hdr)):
        if s in hdr[n]:
            return n

    return -1


# ==============================================================================
# Find a key in the header and replace its value, or create it if nonexistent.
# ==============================================================================

def header_set(hdr, key, value):

    start = key + ':'
    found = False

    for n in range(0, len(hdr)):
        if hdr[n].startswith(start):
            hdr[n] = start + ' ' + value
            found = True

    if not found:
        hdr += [start + ' ' + value]


# ==============================================================================
# Return the content-length field or -1 if not found.
# ==============================================================================

def header_content_length(hdr):

    idx = header_find(hdr, 'Content-Length:')

    if idx < 0:
        return -1

    try:
        return int(hdr[idx][len('Content-Length:'):])
    except ValueError:
        logging.debug('header_content_length(): Bad Content-Length: ' + hdr[idx])
        return -1


# ==============================================================================
# Interpret the header and open a socket/SSL to the execution host's proxy.
# ==============================================================================

def connect_server(hdr, cache_writer, cache, remoteip):

    arr = hdr[0].split(' ')

    if len(arr) < 2:
        if hdr[0] != '':
            fatal('Malformed header2: ' + hdr[0])
        else:
            logging.debug('Malformed header2: ' + hdr[0])
            log_header(hdr)
            fatal("Can't go on")

    url = arr[1]

    if url.find('?') != -1:
        arr = url.split('?')
        # if len(arr) != 2:
        #  fatal("Did not expect two '?' in URL")
        url = arr[0]

    comp = url.split('/')[1:]

    if cfg.VERBOSITY > 2:
        log_header(hdr)

    if comp[0] not in ['weber', 'notebook', 'health']:
        fatal('Malformed URL: ' + url)

    if comp[0] == 'health':
        return None, None, "health"

    try:
        sessnum = int(comp[1])
    except BaseException:
        fatal('Malformed URL: ' + url)

    cookie = comp[2]

    try:
        ct = int(comp[3])
    except BaseException:
        fatal('Malformed URL: ' + url)

    if header_find(hdr, 'Upgrade: websocket') == -1:
        #  header_set(hdr, 'Connection', 'close')
        pass

    if sessnum in cache:
        #logging.debug('Cache hit')
        values = cache[sessnum]
    else:
        #logging.debug('Cache miss')
        values = find_session_params(sessnum, remoteip)
        if values[0] == '':
            # Cache the invalid entry so we don't continually
            # bother the database for invalid sessions.
            write_cache_pipe(cache_writer, sessnum, '', '', '', -1)
            fatal('Session %d not found. rhost=%s' % (sessnum, remoteip))

        if values[1] != cookie:
            fatal('Session cookie does not match. rhost=%s: [%s][%s]' % (remoteip, values[1], cookie))

        if values[3] != ct:
            fatal('Session container does not match. rhost=%s' % remoteip)

        write_cache_pipe(cache_writer, sessnum, values[0], values[1], values[2], values[3])

        cache[sessnum] = values

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
   
    if cfg.FORWARD_SSL:
        if hasattr(ssl, "TLSv1_2_METHOD"):
            ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        else:
            ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1)

        sock = ctx.wrap_socket(sock, server_hostname = values[2])

    try:
        sock.connect((values[2], cfg.FORWARD_PORT))
        #send_header(sock, hdr, body)
        return sock, sessnum, values[0]
    except BaseException:
        fatal("Unable to forward to %s:%d" % (values[2], cfg.FORWARD_PORT))


#==============================================================================
# Establish a socket to the target address.
#==============================================================================

def connect_container(hdr):
    port = cfg.FORWARD_PORT
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    host = None

    arr = hdr[0].split(' ')

    if len(arr) < 2:
      fatal('Malformed header2: ' + hdr[0])

    url = arr[1]

    if url.find('?') != -1:
        arr = url.split('?')

        if len(arr) != 2:
            fatal("Did not expect two '?' in URL")

        url = arr[0]

    comp=url.split('/')[1:]

    if cfg.VERBOSITY > 0:
        logging.info(url)

    if len(comp) < 1 or comp[0] not in [ 'weber', 'notebook' ]:
        fatal('Malformed URL: (comp0) ' + url)
    try:
        ct=int(comp[3])
    except:
        fatal('Malformed URL: (comp3) ' + url)

    if cfg.VIRT_ENGINE == 'openvz':
        try:
            machine_number = int(socket.gethostbyname(socket.getfqdn()).split('.')[3])
            digit = ct/256 + (machine_number % 64) * 4
            host = cfg.PRIVATE_NET % (digit % 256, ct %256)
        except:
            fatal('Unable to determine machine number')

    elif cfg.VIRT_ENGINE == 'docker':
        try:
            if cfg.DOCKERTOOL_NET == '0.0.0.0':
                fatal('DOCKERTOOL_NET configuration variable is undefined, please edit the configuration file for the exec-proxy service')
            gwbytes = cfg.DOCKERTOOL_NET.split('.')
            IPa = gwbytes[0:2] + [str(int(gwbytes[2]) + ct/256), str(ct %256)]
            logging.debug('tool container IP is ' + ".".join(IPa))
            host = ".".join(IPa)
        except:
            fatal('Unable to determine tool network IP from ' + str(ct))

    else:
        fatal('VIRT_ENGINE has an unsupported value, please edit the configuration file')

    for x in range(1,80):
        try:
            sock.connect((host, port))

            if x > 10:  # If less than a second, don't issue a warning.
                logging.debug("Made %d attempts to connect before success." % x)

            #n = header_find(hdr, 'Connection:')
            #if n >= 0:
            #  log("'%s'" % hdr[n])

            return sock
        except:
            time.sleep(0.1 * x)

    fatal("Unable to forward to %s:%d" % (host,port))
    
    return False


# ==============================================================================
# Copy a plaintext socket or SSL socket bidirectionally.
# ==============================================================================

def bidirectional_copy(a, b):

    ahost, aport = a.getpeername()
    bhost, bport = b.getpeername()

    logging.debug("bidirectionl_copy(%s:%d <-> %s:%d)" % (ahost, aport, bhost, bport))

    setblocking(a, False)
    setblocking(b, False)

    af = a.fileno()
    bf = b.fileno()

    amsg = b''
    bmsg = b''

    rfds = [af, bf]
    wfds = []

    while len(rfds) > 0 or len(wfds) > 0:

        try:
            rd, wr, _ = select.select(rfds, wfds, [])
        except select.error as e:
            if e.errno == errno.EAGAIN:
                logging.debug("bidirectionl_copy(%s:%d <-> %s:%d): EAGAIN (continue)" % (ahost, aport, bhost, bport))
                continue
            elif e.errno == errno.EBADF:
                logging.debug("bidirectionl_copy(%s:%d <-> %s:%d): EBADF (shutdown)" % (ahost, aport, bhost, bport))
                break
            else:
                logging.debug("bidirectionl_copy(%s:%d <-> %s:%d): errno = %d (shutdown)" % (ahost, aport, bhost, bport, e.errno))
                break

        if af in rd:
            try:
                amsg += read_chunk(a)

                if bf not in wfds:
                    wfds += [bf]
                    wr += [bf]

            except ReaderClose:
                logging.debug("bidirectionl_copy(%s:%d <-> %s:%d): source closed (drain)" % (ahost, aport, bhost, bport))
                rfds.remove(af)

                if bf not in wfds:
                    wfds += [bf]

                continue

        if bf in rd:
            try:
                bmsg += read_chunk(b)

                if af not in wfds:
                    wfds += [af]
                    wr += [af]

            except ReaderClose:
                logging.debug("bidirectionl_copy(%s:%d <-> %s:%d): destination closed (drain)" % (ahost, aport, bhost, bport))
                rfds.remove(bf)

                if af not in wfds:
                    wfds += [af]

                continue

        if bf in wr:
            if len(amsg) > 0:
                amsg = write_chunk(b, amsg)

            if len(amsg) == 0:
                wfds.remove(bf)

        if af in wr:
            if len(bmsg) > 0:
                bmsg = write_chunk(a, bmsg)

            if len(bmsg) == 0:
                wfds.remove(af)

        if af not in rfds and len(amsg) == 0:
            #shutdown(b)

            if bf in rfds:
                rfds.remove(bf)

        if bf not in rfds and len(bmsg) == 0:
            #shutdown(a)

            if af in rfds:
                rfds.remove(af)


# ==============================================================================
# Forward content of a particular length from a to b.
# ==============================================================================

def forward_body(a, b, blen):
    setblocking(a, True)
    setblocking(b, True)

    # Read the stream until it's closed.

    if blen == -1:
        while True:
          try:
              chunk = read_chunk(a,4096)
          except ReaderClose:
              return

          while len(chunk) > 0:
              chunk = write_chunk(b, chunk)

    # Read the stream for the precise number of bytes

    while blen > 0:
        try:
            chunk = read_chunk(a, min(blen, 4096))
        except ReaderClose:
            return

        blen -= len(chunk)

        while len(chunk) > 0:
            chunk = write_chunk(b, chunk)



# ==============================================================================
# Look up the username, filexfer cookie, execution host, and container number.
# ==============================================================================

def find_session_params(sessnum, remoteip):

    db = MySQLdb.connect(host=cfg.MYSQL_HOST, user=cfg.MYSQL_USER,
                         passwd=cfg.MYSQL_PASSWORD, db=cfg.MYSQL_DB)

    c = db.cursor()

    c.execute("""SELECT fileperm.fileuser, fileperm.cookie, display.hostname, display.dispnum
        FROM fileperm, display
        WHERE fileperm.sessnum = display.sessnum
        AND display.sessnum = %d""" % sessnum, [])

    rows = c.fetchall()

    if len(rows) == 1:
        user, cookie, host, ct = rows[0]
    else:
        user, cookie, host, ct = '', '', '', -1

    db.close()

    return user, cookie, host, ct


# ==============================================================================
# Add an entry into the view table for an active WebSocket connection.
# ==============================================================================

def add_viewer(sessnum, username, host):
    logging.debug("add_viewer(%s, %s, %s)" % (sessnum, username, host))

    db = MySQLdb.connect(host=cfg.MYSQL_HOST, user=cfg.MYSQL_USER, passwd=cfg.MYSQL_PASSWORD, db=cfg.MYSQL_DB)

    c = db.cursor()

    c.execute("""SELECT timeout FROM session WHERE sessnum=%s""", [sessnum])

    row = c.fetchone()

    if row is None:
        timeout = None
    else:
        timeout = row[0]

    if timeout is None:
        fatal('add_viewer(): Unable to get timeout after insert')

    c.execute("""INSERT INTO view(sessnum,username,remoteip,start,heartbeat)
        VALUES(%s, %s, %s, NOW(), NOW())""", [sessnum, username, host])

    c.execute("""SELECT last_insert_id()""", [])

    row = c.fetchone()

    db.close()

    if row is None:
        view_str = None
    else:
        view_str = row[0]

    if view_str is None:
        fatal("add_viewer(): Unable to get view ID after insert")

    signal.signal(signal.SIGALRM, heartbeat_handler)
    signal.alarm(heartbeat_interval)

    return int(view_str), int(timeout / 4)


# ==============================================================================
# Remove the view entry when the WebSocket is closed.
# ==============================================================================

def delete_viewer(viewid):
    global heartbeat_viewid

    logging.debug("delete_viewer(%s)" % (viewid))

    if (viewid != int(viewid)) or (viewid <= 0):
        logging.error("delete_viewer(%s): invalid view id" % viewid)
        return

    db = MySQLdb.connect(host=cfg.MYSQL_HOST, user=cfg.MYSQL_USER, passwd=cfg.MYSQL_PASSWORD, db=cfg.MYSQL_DB)

    c = db.cursor()

    c.execute("""INSERT INTO viewlog (sessnum,username,remoteip,time,duration)
       SELECT view.sessnum, view.username, view.remoteip, view.start, TIMESTAMPDIFF(SECOND, start, NOW())
       FROM view WHERE view.viewid = %d""" % viewid, [])

    db.close()

    signal.alarm(0)
    signal.signal(signal.SIGALRM, signal.SIG_DFL)

    heartbeat_viewid = 0


# ==============================================================================
# This function is invoked by a signal and updates the view table entry to
# indicate that a connection is still in progress.
# ==============================================================================

def heartbeat_handler(signo, extra):

    logging.debug("heartbeat_handler(%s)" % (heartbeat_viewid))
    
    db = MySQLdb.connect(host=cfg.MYSQL_HOST, user=cfg.MYSQL_USER, passwd=cfg.MYSQL_PASSWORD, db=cfg.MYSQL_DB)

    c = db.cursor()

    c.execute("""UPDATE view SET heartbeat=NOW() WHERE viewid=%d""" % heartbeat_viewid, [])

    db.close()

    signal.alarm(heartbeat_interval)


# ==============================================================================
# Don't let the connection read too long without a full header.
# ==============================================================================

def timeout_handler(signo, extra):

    raise Timeout

# ==============================================================================
# Handle a new connection.
# ==============================================================================

def handle_front_proxy_connection(ns, ctx, cache_writer, cache, remoteip, remoteport):
    global heartbeat_interval
    global heartbeat_viewid

    if ctx != None:
        try:
            logging.debug("handle_front_proxy_connection(): initiating ssl session")
            ns = ctx.wrap_socket(ns, server_side = True)
        except ssl.SSLError as errors:
            if 'wrong version number' in str(errors):
                logging.info("handle_front_proxy_connection(): invalid ssl handshake")
            elif 'Unexpected EOF' not in str(errors) and 'ECONNRESET' not in str(errors):
                logging.info("handle_front_proxy_connection(): Caught ssl.SSLError: %s (close connection)" % str(errors))
            else:
                logging.info("handle_front_proxy_connection(): Caught ssl.SSLError: %s (close connection)" % str(errors))
            return
        except OSError as errors:
            logging.info("handle_front_proxy_connection(): Caught OSError: %s (close connection)" % str(errors))
            return

    host, port = ns.getpeername()
    logging.debug("%s:%d Connected" % (host,port))

    heartbeat_interval = cfg.HEARTBEAT_INTERVAL
    newconn = None
    status = 0

    while True:

        try:
            hdr, body = read_header(ns, cfg.HEADER_TIMEOUT)
        except Timeout:
            status = 408
            break

        if len(hdr) < 1 or hdr[0] == '':
            status = 0
            break

        if cfg.LISTEN_SSL:
            # The proxy is the primary front-end web interface.  Create the
            # X-Real-IP and X-Forwarded-For header entries.  Replace others.

            header_set(hdr, 'X-Forwarded-For', remoteip)
            header_set(hdr, 'X-Forwarded-Port', str(cfg.LISTEN_PORT))
            header_set(hdr, 'X-Forwarded-Proto', "https")
            header_set(hdr, 'X-Real-IP', remoteip)
            header_set(hdr, 'X-Scheme', "https")

        else:
            # If not using SSL, assume that the connection is forwarded from
            # Apache or some other kind of front-facing web server that adds an
            # X-Real-IP or X-Forwarded-For entry to the HTTP header to indicate
            # the true origin of the connection.

            idx = header_find(hdr, 'X-Real-IP')

            if idx == -1:
                idx = header_find(hdr, 'X-Forwarded-For')

            if idx != -1:
                arr = hdr[idx].split(':')

                if len(arr) != 2:
                    logging.warning('handle_connection(%s:%d): Malformed X-Real-IP line: %s' % (host, port, hdr[idx]))
                else:
                    remoteip = arr[1].strip()

        logging.info("%s:%d %d %s" % (host, port, header_content_length(hdr), hdr[0]))

        # Set up the forwarded connection to the execution host proxy.
        if (newconn == None):
            newconn, sessnum, username = connect_server(hdr, cache_writer, cache, remoteip)

        if newconn == None and sessnum == None and username == 'health':
            status = 200
            break

        if not newconn:
            status = 502
            break

        send_header(newconn, hdr)
        send_data(newconn, body)

        if header_find(hdr, 'Upgrade: websocket') != -1:
            logging.info("%s:%d Upgrade: websocket" % (host,port))
            # The connection is upgraded to a websocket.  Record the duration of
            # the connection in the 'view' table of the database.

            heartbeat_viewid, heartbeat_interval = add_viewer(sessnum, username, remoteip)

            try:
                bidirectional_copy(ns, newconn)
            except SystemExit:
                pass

            if heartbeat_viewid > 0:
                delete_viewer(heartbeat_viewid)

            break

        else:

            bodylen = header_content_length(hdr)

            if bodylen < 0:
                bodylen = 0

            if bodylen > 0:
                forward_body(ns, newconn, bodylen - len(body))

            hdr, body = read_header(newconn)

            bodylen = header_content_length(hdr)

            logging.info("%s:%d %d %s" % (host,port,bodylen,hdr[0]))

            send_header(ns, hdr)
            send_data(ns, body)
            #header_set(hdr, 'Connection', 'close')


            if bodylen < 0:
                bodylen = 0

            if bodylen > 0:
                forward_body(newconn, ns, bodylen - len(body))

            #send_header(ns, hdr)
            #send_body(ns, body)
            #bidirectional_copy(newconn, ns)

    if (newconn):
        shutdown(newconn)

    if (status == 200):
        hdr = ['HTTP/1.1 200 OK','Connection: Closed']
        
        try:
            send_header(ns, hdr, timeout = cfg.HEADER_TIMEOUT)
        except Timeout:
            pass

        logging.info("%s:%d -1 HTTP/1.1 200 OK" % (host,port))

    if (status == 502):
        hdr = ['HTTP/1.1 502 Bad Gateway','Connection: Closed']
        
        try:
            send_header(ns, hdr, timeout = cfg.HEADER_TIMEOUT)
        except Timeout:
            pass

        logging.info("%s:%d -1 HTTP/1.1 502 Bad Gateway" % (host,port))

    if (status == 408):
        hdr = ['HTTP/1.1 408 Request Timeout','Connection: Closed']
        
        try:
            send_header(ns, hdr, timeout = cfg.HEADER_TIMEOUT)
        except Timeout:
            pass

        logging.info("%s:%d -1 HTTP/1.1 408 Request Timeout" % (host,port))

    shutdown(ns)

    logging.debug("%s:%d Closed" % (host,port))


#==============================================================================
# Handle a new exec proxy connection.
#==============================================================================

def handle_exec_proxy_connection(ns, ctx, remoteip, remoteport):

    if ctx != None:
        try:
            logging.debug("handle_exec_proxy_connection(): initiating ssl session")
            ns = ctx.wrap_socket(ns, server_side = True)
        except ssl.SSLError as errors:
            if 'wrong version number' in str(errors):
                logging.info("handle_exec_proxy_connection(): invalid ssl handshake")
            elif 'Unexpected EOF' not in str(errors) and 'ECONNRESET' not in str(errors):
                logging.info("handle_exec_proxy_connection(): Caught ssl.SSLError: %s (close connection)" % str(errors))
            else:
                logging.info("handle_exec_proxy_connection(): Caught ssl.SSLError: %s (close connection)" % str(errors))
            return
        except OSError as errors:
            logging.info("handle_exec_proxy_connection(): Caught OSError: %s (close connection)" % str(errors))
            return

    host, port = ns.getpeername()
    logging.debug("%s:%d Connected" % (host,port))

    try:
        hdr,body = read_header(ns, cfg.HEADER_TIMEOUT)
    except Timeout:
        fatal('Timeout waiting for header ' + str(debug_header))

    if hdr == '':
        return

    newconn = connect_container(hdr)

    if not newconn:
        # TODO: issue a 404 response or something.
        ns.close()
        return
  
    send_header(newconn,hdr,body)

    bidirectional_copy(ns, newconn)


# ==============================================================================
# Accept a new connection, handle it as a child process, incorporate the
# child's database use into a cache.
# ==============================================================================

def main_loop(ls, ctx):

    session_cache = {}
    rlist = []
    logging.debug("Main loop listening on %d" % ls.fileno())
    while True:

        try:
            logging.debug("main_loop(): entering accept()")
            ns, addr = ls.accept()
        except OSError:
            # accept() could be interrupted by a signal.
            logging.info("accept() OSError exception, retrying")
            continue

        if cfg.MODE == 'front':
            t = time.time()

            read_cache_pipes(rlist, session_cache, t)
        
            r, w = os.pipe()

        pid = os.fork()

        if pid == 0:
            # CHILD PROCESS BEGINS

            os.setsid()  # disassociate from calling process

            if os.fork() == 0:
                # GRANDCHILD PROCESS BEGINS

                signal.signal(signal.SIGINT, child_termination_handler)
                signal.signal(signal.SIGTERM, child_termination_handler)

                ls.close()  # Close the listener in the child

                if cfg.MODE == 'front':
                    os.close(r)  # Close reader side of cache pipe as child is the writer

                try:
                    if cfg.MODE == 'front':
                        handle_front_proxy_connection(ns, ctx, w, session_cache, addr[0], addr[1])
                    else:
                        handle_exec_proxy_connection(ns, ctx, addr[0], addr[1])

                except SystemExit:
                    pass
                except BaseException:
                    log_exception("Exception in handle_connection():")

                # If handle_connection() failed because of an exception, try to do
                # a delete_viewer() here.  This one will not be subject to exception
                # notification, so we won't catch any bugs with this call.  Instead,
                # we rely on the call inside handle_connection() to happen in most
                # circumstances and notify us if it fails.

                if heartbeat_viewid > 0:
                    delete_viewer(heartbeat_viewid)

                # GRANDCHILD PROCESS ENDS
                # This process has init as a parent which will handle its reaping
                os._exit(0)

            # CHILD PROCESS ENDS
            # We will let this process exit so that the grandchild
            # will be reparented to init and not create zombie
            # processes. The parent process will reap with waitpid()
            os._exit(0)

        # PARENT PROCESS CONTINUES

        os.waitpid(pid,0) # wait for child process to end

        if cfg.MODE == 'front':
            # Add the pipe reader to a list of other pipe readers.
            # Make it non-blocking so that a read on a pipe that is not ready
            # will result in EAGAIN.  This allows us to try reading all of the
            # pipes instead of having to select() on them.

            os.close(w)  # Close writer side of cache pipe as parent is the reader

            rlist += [r]

            setblocking(r, False)

        flush_cache(session_cache, t)

        ns.close()

        # REPEAT WHILE LOOP FOR NEXT CONNECTION


# ==============================================================================
# Run the process in the background.
# ==============================================================================

def daemonize():

    os.chdir('/')

    pid = os.fork()

    if pid != 0:
        os.waitpid(pid,0)
        os._exit(0)

    os.setsid()

    pid = os.fork()
    
    if pid != 0:
        os._exit(0)

    # Grandchild of original process returns
    
    # This grandchild has no controlling terminal
    # This grandchild can not unintentionally obtain a controlling terminal
    # This grandchild has a ppid of init
    # This grandchild has a sid of non-existent process
    # This grandchild has a pgid of a non-existent process


# ==============================================================================
# Create the PID file.
# ==============================================================================

def write_pidfile(name):

    f = open(name, 'w')
    f.write(str(os.getpid()) + '\n')
    f.close()


# ==============================================================================
# Remove the PID file.
# ==============================================================================

def remove_pidfile(name):

    if name is None:
        return

    try:
        os.seteuid(0)
        os.unlink(name)
    except BaseException:
        logging.info('remove_pidfile(): Unable to remove %s' % name)


# ==============================================================================
# Clean up on termination.
# ==============================================================================

def child_termination_handler(signo, extra):

    if signo != 0:
        logging.info('Terminating on signal %d' % signo)

    if heartbeat_viewid > 0:
        delete_viewer(heartbeat_viewid)

    # There is no good reason for termination.  Consider it a failure.
    os._exit(1)


def parent_termination_handler(signo, extra):

    if signo != 0:
        logging.info('Terminating on signal %d' % signo)

    remove_pidfile(cfg.PIDFILE)

    # There is no good reason for termination.  Consider it a failure.
    os._exit(1)


# ==============================================================================
# The main function creates the main listening socket, sets up logging and
# daemonization, creates the pid file, invokes main_loop(), and handles
# top-level exceptions.
# ==============================================================================

def main():

    global heartbeat_viewid 
    global heartbeat_interval
    global debug_header
    global cfg

    heartbeat_viewid = 0
    heartbeat_interval = -1
    debug_header = b''
    cfg = Config()

    #FRONT_LISTEN_PORT = 8080      # port number to bind to
    #FRONT_LISTEN_SSL = False      # Listen as SSL socket (True/False)
    #FRONT_KEY = "/etc/ssl/private/ssl-cert-snakeoil.key"  # SSL private key
    #FRONT_CERT = "/etc/ssl/certs/ssl-cert-snakeoil.pem"  # SSL certificate
    #FRONT_CACERT = None       # SSL CA certificate

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", action='append', help="configuration file to use", metavar='CONFIGFILE')
    parser.add_argument("-v", "--verbose", action='count', help="increase verbosity level", default=0)
    parser.add_argument("-f", "--foreground", action='store_true', help="run in foreground", default=False)

    args = parser.parse_args()

    if args.config:
        for config in args.config:
            cfg.read(config)

    if args.foreground:
        cfg.FOREGROUND = True

    if args.verbose:
        cfg.VERBOSITY = args.verbose

    # we're root so write to syslog until we change uid instead of writing as root in a directory owned by a different user
    syslog.syslog("weber starting up")

    if cfg.LISTEN_SSL:
        ctx = create_ssl_context()
    else:
        ctx = None

    # Create the listening socket before backgrounding or writing the pidfile.
    try:
        ls = create_listener(cfg.LISTEN_ADDRESS, cfg.LISTEN_PORT)
    except Exception as e:
        syslog.syslog("FATAL: Unable to listen to %s:%d" % (cfg.LISTEN_ADDRESS, cfg.LISTEN_PORT))
        syslog.syslog("FATAL: Exception is " + str(e))
        print("FATAL: Exception is " + str(e))
        raise
        os._exit(1)

    if not cfg.FOREGROUND:
        try:
            fd = os.open("/dev/null", os.O_RDWR)
            os.dup2(fd, 0)
            os.dup2(fd, 1)
            os.dup2(fd, 2)
            os.close(fd)
        except Exception as e:
            syslog.syslog("FATAL: Unable to use /dev/null for input/output")
            syslog.syslog("FATAL: Exception is " + str(e))
            print("FATAL: Unable to use /dev/null for input/output")
            os._exit(1)

    # If not set to run as a 'foreground' application we 
    # daemonize
    #
    
    if not cfg.FOREGROUND:
        daemonize()

    write_pidfile(cfg.PIDFILE)

    signal.signal(signal.SIGINT, parent_termination_handler)
    signal.signal(signal.SIGTERM, parent_termination_handler)

    if os.getuid() != 0:
        if cfg.RUN_UID is not None:
            fatal('Not running as root, unable to change uid to %s' % cfg.RUN_UID)
        if cfg.RUN_GID is not None:
            fatal('Not running as root, unable to change gid to %s' % cfg.RUN_GID)
    else:
        if cfg.RUN_UID is None:
            fatal('Running as root, RUN_UID not set, unable to change uid')

    if cfg.RUN_GID is not None:
        try:
            os.setegid(grp.getgrnam(cfg.RUN_GID).gr_gid)
        except BaseException:
            fatal('Unable to set gid to %s' % cfg.RUN_GID)

    if cfg.RUN_UID is not None:
        try:
            os.seteuid(pwd.getpwnam(cfg.RUN_UID).pw_uid)
        except BaseException:
            fatal('Unable to set uid to %s' % cfg.RUN_UID)

    logging.Formatter.converter = time.gmtime

    if cfg.VERBOSITY > 0:
        loglevel = logging.INFO
    elif cfg.VERBOSITY > 10:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.WARNING

    logging.basicConfig(level=loglevel,
                        format='%(asctime)s.%(msecs)03dZ %(process)5d %(levelname)-8s %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%S',
                        filename=cfg.LOGFILE,
                        filemode='a')

    logging.info('Starting up')

    if cfg.VERBOSITY > 0:
        logging.info('Verbosity level %d' % cfg.VERBOSITY)

    # Invoke the handler loop.  Catch and print any unusual exceptions.
    try:
        main_loop(ls, ctx)
    except SystemExit:  # Something called fatal()
        pass
    except BaseException:
        log_exception("Exception in server:")

    # Once it's done, clean up.  A zero indicates no signal.
    parent_termination_handler(0, 0)



# We implement an application config class which can be populated from 
# hubzero middleware classic python config files or with several possible
# configuration file formats (ini,yml,json,env) or environment variables.

class Config:
    _section = 'weber'
    _prefix = 'WEBER_'

    # Map old python config file variable names to a normalized form
    # that can be sensibly prefixed with _prefix to make environment
    # variables suitable for configuration with docker.

    _varmap = {
        'FRONT_PROXY_CONFIG_FILE':  'CONFIG_FILE',
        'FRONT_CONFIGDIR': 'CONFIG_DIR',
        'FRONT_LISTEN_HOST': 'LISTEN_ADDRESS',
        'FRONT_LISTEN_PORT': 'LISTEN_PORT',
        'FRONT_LISTEN_SSL': 'LISTEN_SSL',
        'FRONT_LISTEN_TLSv1_2': 'LISTEN_TLSv1_2',
        'FRONT_KEY': 'SSL_PRIVKEY',
        'FRONT_CERT': 'SSL_CERT',
        'FRONT_CACERT': 'SSL_CACERT',
        'FRONT_DHPARAM': 'SSL_DHPARAM',
        'FRONT_CIPHERS': 'SSL_CIPHERS',
        'FORWARD_HOST': 'FORWARD_ADDRESS',
        'FORWARD_PORT': 'FORWARD_PORT',
        'FORWARD_SSL': 'FORWARD_SSL',
        'FORWARD_TLSv1_2': 'FORWARD_TLSv1_2',
        'PIDFILE': 'PIDFILE',
        'PROXY_LOG': 'LOGFILE',
        'HEADER_TIMEOUT': 'HEADER_TIMEOUT',
        'CACHE_FLUSH_PERCENT': 'CACHE_FLUSH_PERCENT',
        'CACHE_DURATION': 'CACHE_DURATION',
        'RUN_UID': 'RUN_UID',
        'RUN_GID': 'RUN_GID',
        'HEARTBEAT_INTERVAL': 'HEARTBEAT_INTERVAL',
        'VERBOSITY': 'VERBOSITY',
        'FOREGROUND': 'FOREGROUND',
        'EXEC_PROXY_CONFIG_FILE': 'CONFIG_FILE',
        'EXEC_CONFIGDIR': 'CONFIG_DIR',
        'EXEC_LISTEN_HOST': 'LISTEN_ADDRESS',
        'EXEC_LISTEN_PORT': 'LISTEN_PORT',
        'EXEC_LISTEN_SSL': 'LISTEN_SSL',
        'EXEC_LISTEN_TLSv1_2': 'LISTEN_TLSv1_2',
        'EXEC_SERVER_KEY': 'SSL_PRIVKEY',
        'EXEC_SERVER_CERT': 'SSL_CERT',
        'EXEC_SERVER_CACERT': 'SSL_CACERT',
        'EXEC_SERVER_DHPARAM': 'SSL_DHPARAM',
        'EXEC_SERVER_CIPHERS': 'SSL_CIPHERS',
        'VIRT_ENGINE': 'VIRT_ENGINE',
        'DOCKERTOOL_NET': 'DOCKERTOOL_NET',
        'PRIVATE_NET': 'PRIVATE_NET',
        'WEBER_MODE': 'MODE',
        'mysql_port': 'MYSQL_PORT',
        'mysql_host': 'MYSQL_HOST',
        'mysql_user': 'MYSQL_USER',
        'mysql_password': 'MYSQL_PASSWORD',
        'mysql_db': 'MYSQL_DB'
    }

    def __init__(self):
        self.CONFIG_FILE = '/etc/mw-proxy/front-proxy.conf'
        self.CONFIG_DIR = '/etc/mw-proxy'
        self.LISTEN_ADDRESS = "0.0.0.0"  # IP address to bind to
        self.LISTEN_PORT = 8888      # port number to bind to
        self.LISTEN_SSL = True      # Listen as SSL socket (True/False)
        self.LISTEN_TLSv1_2 = False  # Newer TLS version requiring Python 2.7.9 at least
        self.SSL_PRIVKEY = "/etc/mw-proxy/front-proxy.key"
        self.SSL_CERT = "/etc/mw-proxy/front-proxy.cert"
        self.SSL_CACERT = '/etc/mw-proxy/front-proxy.ca-bundle'
        self.SSL_DHPARAM = None      # Diffie-Hellman param
        self.SSL_CIPHERS = None      # List of ciphers to use/exclude
        self.FORWARD_ADDRESS = ''
        self.FORWARD_PORT = 443        # Port to connect to on execution hosts
        self.FORWARD_SSL = True        # Use SSL to connect to execution hosts (True/False)
        self.FORWARD_TLSv1_2 = True    # Use TLS 1.2 to connect to execution hosts (True/False)
        self.PIDFILE = '/var/run/front-proxy.pid'
        self.LOGFILE = '/var/log/front-proxy/front-proxy.log'
        self.HEADER_TIMEOUT = 20       # How long should we wait for an HTTP header
        self.CACHE_FLUSH_PERCENT = 1     # How many times per hundred accesses is cache flushed
        self.CACHE_DURATION = 86400      # Max age of something to keep in cache
        self.RUN_UID = 'hz-front-proxy'  # UID to run the proxy
        self.RUN_GID = 'hz-front-proxy'  # GID to run the proxy
        self.HEARTBEAT_INTERVAL = 3600   # The default heartbeat interval is 1 hour.
        self.VERBOSITY = 0
        self.FOREGROUND = False
        self.VIRT_ENGINE = 'openvz'
        self.PRIVATE_NET = '0.0.0.0'
        self.DOCKERTOOL_NET = '0.0.0.0'
        self.MODE = 'front'
        self.MYSQL_USER = 'root'
        self.MYSQL_PASSWORD = 'password'
        self.MYSQL_HOST = 'localhost'
        self.MYSQL_PORT = None
        self.MYSQL_DB = 'narwhal'


    def __str__(self):
        output = '{'

        for key,value in vars(self).items():

            if key.startswith('_'):
                continue
            if key == 'read':
                continue

            if output != '{':
                output += ','

            output += "'" + str(key) + "': '" + str(value) + "'"

        output += '}'

        return output

    def _dequote(self, s):

        ss = s.strip()

        if (ss[0] == '\''and ss[-1:] == '\'') or (ss[0] == '"' and ss[-1:] == '"'):
            return ss[1:-1]

        return s

    def read(self, filename = None, type = 'py'):
    
        map = False  
        prefix = False
        ext = None

        if filename != None:
            _,ext = os.path.splitext(filename)

            ext = ext[1:].lower()

    
        if (ext == 'py' or ext == 'conf' ):
            myglobals = {}
            config = {}

            with open(filename) as fh:
                pycode = fh.read()

                exec(pycode, myglobals, config)

                try:
                    config['PRIVATE_NET'] = config['CONTAINER_CONF']['PRIVATE_NET']
                except KeyError:
                    pass

            map = True

        elif (ext == 'ini'):
            config = ConfigParser()
            config.optionxform = str
            config.read(filename)

            config = dict(config.items(self._section))
            config = {k:self._dequote(v) for k,v in config.items()}

        elif (ext == 'json'):
            with open(filename) as fh:
                config = simplejson.load(fh)

            config = config[self._section]

        elif (ext == 'yml' or ext == 'yaml'):
            with open(filename) as fh:
                config = yaml.load(fh)

            config = config[self._section]

        elif (ext == 'env'):
            config = {}

            with open(filename) as fh:
                for line in fh:
                    line = line.strip()

                    if line.startswith('#'):
                        continue
         
                    key, value = line.split('=')

                    if key.startswith('export '):
                        key = key[7:]

                    config[key] = self._dequote(value)

            prefix = True

        elif (filename == None):
            config = os.environ
            prefix = True
    
        for key in config:
            value = config[key]

            if (prefix):
                if key.startswith(self._prefix):
                    key = key[len(self._prefix):]

            if (map):
                if key in self._varmap:
                    key = self._varmap[key]

            if (hasattr(self, key)):
                setattr(self,key,value)
        

# ==============================================================================
# Invoke the main function.
# ==============================================================================

if __name__ == "__main__":
    main()
